"""Tests for config package."""
import os
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest

from detective import config


def test_find_hass_config():
    """Test finding HASS config."""
    with patch.dict(os.environ, {"HASSIO_TOKEN": "yo"}):
        assert config.find_hass_config() == "/config"

    with patch.dict(os.environ, {}, clear=True), patch(
        "detective.config.default_hass_config_dir", return_value="default-dir"
    ), patch("os.path.isdir", return_value=True):
        assert config.find_hass_config() == "default-dir"

    with patch.dict(os.environ, {}, clear=True), patch(
        "os.path.isdir", return_value=False
    ), pytest.raises(ValueError):
        config.find_hass_config()


def test_load_hass_config():
    """Test loading hass config."""
    with tempfile.TemporaryDirectory() as tmpdir:
        p = Path(tmpdir)
        with (p / "configuration.yaml").open("wt") as fp:
            fp.write(
                """
mock_secret: !secret some_secret
included: !include included.yaml
nested_with_secret: !include includes/with_secret.yaml
nested_without_secret: !include includes/without_secret.yaml
mock_env: !env_var MOCK_ENV
mock_env: !env_var MOCK_ENV
mock_dir_list: !include_dir_list ./zxc
mock_dir_merge_list: !include_dir_merge_list ./zxc
mock_dir_named: !include_dir_named ./zxc
mock_dir_merge: !include_dir_merge_named ./zxc
# Trigger duplicate error
mock_secret: !secret other_secret
        """
            )

        with (p / "secrets.yaml").open("wt") as fp:
            fp.write(
                """
some_secret: test-some-secret
other_secret: test-other-secret
another_secret: test-another-secret
        """
            )
        with (p / "included.yaml").open("wt") as fp:
            fp.write("some: value")
        includes_dir = p / "includes"
        includes_dir.mkdir(exist_ok=True)
        with (includes_dir / "with_secret.yaml").open("wt") as fp:
            fp.write("some: !secret another_secret")
        with (includes_dir / "without_secret.yaml").open("wt") as fp:
            fp.write("some: value")
        configuration = config.load_hass_config(tmpdir)

    # assert configuration["mock_secret"] == "test-other-secret"  # TODO: fix
    assert configuration["included"] == {"some": "value"}
    assert configuration["nested_with_secret"] == {"some": "test-another-secret"}
    assert configuration["nested_without_secret"] == {"some": "value"}


def test_db_url_from_hass_config():
    """Test extracting recorder url from config."""
    with patch("detective.config.load_hass_config", return_value={}), patch(
        "os.path.isfile", return_value=False
    ), pytest.raises(ValueError):
        config.db_url_from_hass_config("mock-path")

    with patch("detective.config.load_hass_config", return_value={}), patch(
        "os.path.isfile", return_value=True
    ):
        assert (
            config.db_url_from_hass_config("mock-path")
            == "sqlite:///mock-path/home-assistant_v2.db"
        )

    with patch(
        "detective.config.load_hass_config",
        return_value={"recorder": {"db_url": "mock-url"}},
    ):
        assert config.db_url_from_hass_config("mock-path") == "mock-url"



"""Tests for helper functions."""
import pandas as pd

from detective.functions import format_dataframe


def test_format_dataframe_converts_types_and_drops_invalid_rows():
    df = pd.DataFrame(
        {
            "state": ["42.5", "unavailable", "7"],
            "last_updated_ts": [0, 60, None],
            "entity_id": ["sensor.temp", "sensor.temp", "sensor.temp"],
        }
    )

    result = format_dataframe(df)

    assert len(result) == 1
    assert result.iloc[0]["state"] == 42.5
    assert result.iloc[0]["last_updated_ts"] == pd.Timestamp("1970-01-01 00:00:00")
    assert pd.api.types.is_float_dtype(result["state"])
    assert pd.api.types.is_datetime64_any_dtype(result["last_updated_ts"])


def test_format_dataframe_removes_timezone_information():
    df = pd.DataFrame(
        {
            "state": ["1"],
            "last_updated_ts": [1700000000],
        }
    )

    result = format_dataframe(df)

    assert result.iloc[0]["last_updated_ts"].tzinfo is None



"""Tests for auth package."""
import json
from unittest.mock import mock_open, patch

from detective.auth import Auth


def test_auth_from_hass_config():
    """Test extracting users from hass config."""
    mopen = mock_open(read_data=json.dumps({
        'key': 'auth',
        'version': 1,
        'data': {
            'users': [{
                'id': 'mock-user-id',
                'name': 'mock-user-name'
            }],
            'refresh_tokens': [{
                'id': 'mock-refresh-id',
                'user_id': 'mock-user-id',
                'client_name': 'mock-client-name',
                'client_id': 'mock-client-id',
            }]
        }
    }))

    with patch('detective.auth.open', mopen):
        auth = Auth('/some/path')

    assert auth.users == {
        'mock-user-id': {
            'id': 'mock-user-id',
            'name': 'mock-user-name'
        }
    }
    assert auth.refresh_tokens == {
        'mock-refresh-id': {
            'id': 'mock-refresh-id',
            'user': auth.users['mock-user-id'],
            'client_id': 'mock-client-id',
            'client_name': 'mock-client-name',
        }
    }



"""
Helper functions.
"""

import json

import pandas as pd


def format_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """Convert states to numeric where possible and format the last_updated_ts."""
    df["state"] = pd.to_numeric(df["state"], errors="coerce")

    df["last_updated_ts"] = pd.to_datetime(
        df["last_updated_ts"], unit="s", utc=True
    ).dt.tz_convert(None)

    df = df.dropna()
    return df



from unittest.mock import patch, MagicMock
import pandas as pd

from detective.core import get_db_type, stripped_db_url, HassDatabase


def test_get_db_type():
    assert get_db_type("mysql://localhost") == "mysql"
    assert get_db_type("mysql+pymysql://localhost") == "mysql"


def test_stripped_db_url():
    assert stripped_db_url("mysql://localhost") == "mysql://localhost"
    assert stripped_db_url("mysql://paulus@localhost") == "mysql://paulus@localhost"
    assert (
        stripped_db_url("mysql://paulus:password@localhost")
        == "mysql://paulus:***@localhost"
    )


def test_fetch_entities(mock_db):
    with patch.object(
        mock_db,
        "perform_query",
        return_value=[["light.kitchen"], ["light.living_room"], ["switch.ac"],],
    ):
        mock_db.fetch_entities()

    assert mock_db.entities == ["light.kitchen", "light.living_room", "switch.ac"]


mock_data = pd.DataFrame({
    "state": ["20.2", "50.1"],
    "last_updated_ts": ["2025-10-01 12:00:00", "2025-10-01 12:05:00"],
    "entity_id": ["sensor.temperature", "sensor.humidity"],
    "shared_attrs": ["{}", "{}"]
})

@patch("sqlalchemy.create_engine")
def test_fetch_all_sensor_data(mock_create_engine):
    """Test that fetch_all_sensor_data returns a Pandas DataFrame."""
    mock_con = MagicMock()
    mock_engine = MagicMock()
    mock_engine.connect.return_value = mock_con
    mock_create_engine.return_value = mock_engine

    # fetch_all_sensor_data uses read_sql_query under the hood
    with patch("pandas.read_sql_query") as mock_read_sql:
        mock_read_sql.return_value = mock_data

        db = HassDatabase(url="sqlite:///tests/test.db", fetch_entities=False)
        db.con = mock_con

        result = db.fetch_all_sensor_data()
        assert isinstance(result, pd.DataFrame)



"""Tests for time helper."""
from datetime import datetime
from time import time as builtin_time

import pytest

from detective import time


def test_sqlalch_datetime():
    # a datetime string should return a datetime
    result = time.sqlalch_datetime('2018-12-05 10:36:44.900394')
    assert result.year == 2018
    assert result.month == 12
    assert result.day == 5
    assert result.hour == 10
    assert result.minute == 36
    assert result.second == 44
    assert result.microsecond == 900394
    assert result.tzinfo is time.UTC
    
    # a datetime should return a datetime
    result2 = time.sqlalch_datetime(result)
    assert result2 == result
    
    with pytest.raises(ValueError):
        result = time.sqlalch_datetime('garbage')


def test_localize():
    now_time = builtin_time()
    now = datetime.fromtimestamp(now_time)
    utcnow = datetime.fromtimestamp(now_time, time.UTC)
    # There is always a little offset because of how LOCAL_UTC_OFFSET is
    # calculated. We calculate here it's not more than 1 second off
    assert -1 < (time.localize(utcnow) - now).total_seconds() < 1



import detective.core as detective
import detective.functions as functions
import pandas as pd

db_url = "sqlite:///tests/test.db"

def test_db():
    db = detective.HassDatabase(db_url)
    assert db is not None
    assert len(db.entities) > 0

    df = db.fetch_all_data_of(('zone.home',))
    assert df is not None

    df = db.fetch_all_sensor_data(limit=100000)
    assert df is not None

    df = db.fetch_all_statistics_of(("sensor.kitchen", "sensor.living_room", "sensor.ac"), limit=100000)
    assert df is not None
